#include "correlation.h"
#include "img_util.h"

/************************ TO DO 3 **************************/
/*
 *	INPUT: 
 *		origImg:		the original image, 
 *		imgWidth:		the width of the image
 *		imgHeight:		the height of the image
 *						the image is arranged such that 
 *						origImg[3*(row*imgWidth+column)+0], 
 *						origImg[3*(row*imgWidth+column)+1], 
 *						origImg[3*(row*imgWidth+column)+2]
 *						are R, G, B values for pixel at (column, row).
 *
 *      kernel:			the 2D filter kernel,
 *		knlWidth:		the width of the kernel
 *		knlHeight:		the height of the kernel
 *		
 *		scale, offset:  after correlating the kernel with the origImg, 
 *						each pixel should be divided by scale and then added by offset
 *      
 *		selection:      a byte array of the same size as the image, 
 *						indicating where in the original image should be filtered, e.g., 
 *						selection[k] == 1 ==> pixel k should be filtered
 *                      selection[k] == 0 ==> pixel k should NOT be filtered
 *                      a special case is selection is a NULL pointer, which means all the pixels should be filtered. 
 *
 *  OUTPUT:
 *		rsltImg:		the filtered image of the same size as original image.
 *						it is a valid pointer ( allocated already ).
 */

void image_filter(double *rsltImg, const unsigned char *origImg, const unsigned char *selection, 
			int imgWidth, int imgHeight, 
			const double *kernel, int knlWidth, int knlHeight,
			double scale, double offset)
{
	// Confirm memory allocation
	if (rsltImg == NULL)
	{
		printf("ERR: Memory not allocated for resulting image.\n");
		return;
	}

	// For each pixel, apply the kernel at that pixel using pixel_filter
	for (int row = 0;row < imgHeight;++row)
		for (int col = 0; col < imgWidth;++col)
		{
			// Only apply if mask there is 1 or we have no mask
			if ((selection == NULL) || (selection[NINDEX(imgWidth, row, col)] == 1))
			{
				double p[3];
				pixel_filter(p, col, row, origImg, imgWidth, imgHeight, kernel, knlWidth, knlHeight, scale, offset);

				rsltImg[INDEX(imgWidth, row, col, RED)] = p[RED];
				rsltImg[INDEX(imgWidth, row, col, GREEN)] = p[GREEN];
				rsltImg[INDEX(imgWidth, row, col, BLUE)] = p[BLUE];
			}
		}
}

/************************ END OF TBD 3 **************************/


/************************ TO DO 2 **************************/
/*
 *	INPUT: 
 *      x:				a column index,
 *      y:				a row index,
 *		origImg:		the original image, 
 *		imgWidth:		the width of the image
 *		imgHeight:		the height of the image
 *						the image is arranged such that 
 *						origImg[3*(row*imgWidth+column)+0], 
 *						origImg[3*(row*imgWidth+column)+1], 
 *						origImg[3*(row*imgWidth+column)+2]
 *						are R, G, B values for pixel at (column, row).
 *
 *      kernel:			the 2D filter kernel,
 *		knlWidth:		the width of the kernel
 *		knlHeight:		the height of the kernel
 *
 *		scale, offset:  after correlating the kernel with the origImg, 
 *						the result pixel should be divided by scale and then added by offset
 *
 *  OUTPUT:
 *		rsltPixel[0], rsltPixel[1], rsltPixel[2]:		
 *						the filtered pixel R, G, B values at row y , column x;
 */

void pixel_filter(double rsltPixel[3], int x, int y, const unsigned char *origImg, int imgWidth, int imgHeight, 
			  const double *kernel, int knlWidth, int knlHeight,
			  double scale, double offset)
{
	int halfkw = (knlWidth - 1) / 2;
	int halfkh = (knlHeight - 1) / 2;
	int current_kp = -1;					// Current kernel position

	// Perform cross-correlation at x, y
	// Start at top-left to bottom right, by rows
	double sum[3] = {0.0, 0.0, 0.0};
	for (int row = y - halfkh ; row <= (y + halfkh) ; row++)
		for (int col = x - halfkw ; col <= (x + halfkw) ; col++)		
		{
			++current_kp;	// Increment to current kernel position
			if (!IN_RANGE(imgWidth, imgHeight, row, col))
				continue;

			sum[RED]	+= origImg[INDEX(imgWidth, row, col, RED)] * kernel[current_kp];
			sum[GREEN]	+= origImg[INDEX(imgWidth, row, col, GREEN)] * kernel[current_kp];
			sum[BLUE]	+= origImg[INDEX(imgWidth, row, col, BLUE)] * kernel[current_kp];
		}

	rsltPixel[RED] = sum[RED];
	rsltPixel[GREEN] = sum[GREEN];
	rsltPixel[BLUE] = sum[BLUE];

	rsltPixel[RED] = rsltPixel[RED]/scale + offset;
	rsltPixel[GREEN] = rsltPixel[GREEN]/scale + offset;
	rsltPixel[BLUE] = rsltPixel[BLUE]/scale + offset;
}

/************************ END OF TBD 2 **************************/

