
#include "correlation.h"
#include "iScissor.h"
#include "img_util.h"

// two inlined routined that may help;

inline Node &NODE(Node *n, int i, int j, int width)
{ return *(n+j*width+i); }

inline unsigned char PIXEL(const unsigned char *p, int i, int j, int c, int width)
{ return *(p+3*(j*width+i)+c); }

/************************TO DO 1***************************
 *InitNodeBuf
 *	INPUT: 
 *		img:	a RGB image of size imgWidth by imgHeight;
 *		nodes:	an allocated buffer of Nodes of the same size, one node corresponds to a pixel in img;
 *  OUPUT:
 *      initializes the column, row, and linkCost fields of each node in the node buffer.  
 */

//Helper - Returns the length of the given link
double link_length(int link_num)
{
	switch(link_num)
	{
	case 0: 
	case 2:
	case 4:
	case 6:
		return 1.0;
	default:
		return SQRT2;
	}
}

void InitNodeBuf(Node *nodes, const unsigned char *img, int imgWidth, int imgHeight)
{
	if (nodes == NULL)
	{
		printf("ERR: nodes arrays has not been initialized.\n");
		return;
	}

	double maxD = -LARGE_DOUBLE;
	// FIRST PASS - compute row col info for each node and determine all D(link) for each node.  Also determined maxD
	for (int row = 0;row < imgHeight;row++)
		for (int col = 0;col < imgWidth;col++)
		{
			int ncur = NINDEX(imgWidth, row, col);
			int cur = INDEX(imgWidth, row, col, 0);

			nodes[ncur].column = col;
			nodes[ncur].row = row;			
			
			double D[8][3];
			// Compute D for each of the neighbors of (row, col)
			for (int k = 0;k<8;k++)
			{
				// Cross-correlate with the kernel corresponding to the neighbor k and abs-val the result
				pixel_filter(D[k], col, row, img, imgWidth, imgHeight, kernels[k], 3, 3, 1, 0);

				//D[k][RED] = fabs(D[k][RED]);
				//D[k][GREEN] = fabs(D[k][GREEN]);
				//D[k][BLUE] = fabs(D[k][BLUE]);

				nodes[ncur].linkCost[k] = sqrt( (D[k][RED]*D[k][RED] + D[k][GREEN]*D[k][GREEN] + D[k][BLUE]*D[k][BLUE]) );	// Determine D[k] and store in link cost
				nodes[ncur].linkCost[k] > maxD ? maxD = nodes[ncur].linkCost[k] : maxD = maxD;									// Update maximum D value
			}
		}

	// SECOND PASS - use the D values calculated in the first pass to computer link cost
	for (int row = 0;row < imgHeight;row++)
		for (int col = 0;col < imgWidth;col++)
		{
			int ncur = NINDEX(imgWidth, row, col);
			for (int k = 0;k<8;k++)
			{
				nodes[ncur].linkCost[k] = (maxD - nodes[ncur].linkCost[k])*link_length(k);
			}
		}
}

/************************TO DO 4***************************
 *LiveWireDP:
 *	INPUT:
 *		seedX, seedY:	seed position in nodes
 *		nodes:			node buffer of size width by height;
 *      width, height:  dimensions of the node buffer;
 *		selection:		if selection != NULL, search path only in the subset of nodes[j*width+i] if selection[j*width+i] = 1;
 *						otherwise, search in the whole set of nodes. 
 *		numExpanded:		compute the only the first numExpanded number of nodes; (for debugging)
 *	OUTPUT:
 *		computes the minimum path tree from the seed node, by assigning 
 *		the prevNode field of each node to its predecessor along the minimum 
 *		cost path from the seed to that node.
 */

//helper - initialize node states
void init_nodes(Node* nodes, int width, int height, int seedX, int seedY)
{
	for (int k = 0;k<(width*height);k++)
	{
		nodes[k].state = INITIAL;
		nodes[k].prevNode = NULL;
		nodes[k].totalCost = LARGE_DOUBLE;
	}

	nodes[NINDEX(width, seedY, seedX)].totalCost = 0;
	nodes[NINDEX(width, seedY, seedX)].state = ACTIVE;
	
}

void LiveWireDP(int seedX, int seedY, Node *nodes, int width, int height, const unsigned char *selection, int numExpanded)
{
	// Initialize
	int curExpanded = 0;
	CTypedPtrHeap<Node> node_pq;
	init_nodes(nodes, width, height, seedX, seedY);
	node_pq.Insert(&nodes[NINDEX(width, seedY, seedX)]);

	//printf("numExpanded: %d\n", numExpanded);
	// while we are within the tolerance of expansions and our priority queue is not empty
	while ((curExpanded <= numExpanded) && (!node_pq.IsEmpty()))
	{
		// Selects the smallest valuea and marks is fixed
		Node* cur_exp = node_pq.ExtractMin();
		cur_exp->state = EXPANDED;
		++curExpanded;

		//For each of the neighbors, update their total cost with this cost to get to that node
		//from this current node. 
		for (int k = 0;k<8;k++)
		{
			int co, ro;
			int r, c, i;

			r = cur_exp->row; c = cur_exp->column;
			cur_exp->nbrNodeOffset(co, ro, k);
			if (!IN_RANGE(width, height, r+ro, c+co))
				continue;
			
			i = NINDEX(width, r+ro, c+co);			

			if (nodes[i].state != EXPANDED)  // If neighbor already expanded, is fixed
			{
				if (nodes[i].state == INITIAL)	// If not touched before, set the cost and add to queue
				{
					nodes[i].state = ACTIVE;
					nodes[i].totalCost = cur_exp->totalCost + cur_exp->linkCost[k];
					nodes[i].prevNode = cur_exp;
					node_pq.Insert(&nodes[i]);
				}
				else 
				{
					// If another active node, check if new cost is smaller than the one already there
					if ((cur_exp->totalCost + cur_exp->linkCost[k]) < nodes[i].totalCost)
					{
						nodes[i].totalCost = cur_exp->totalCost + cur_exp->linkCost[k];
						nodes[i].prevNode = cur_exp;
						node_pq.Update(&nodes[i]);
					}
				}
			}
		}
	}
}

/************************TO DO 5***************************
 *MinimumPath:
 *	INPUT:
 *		nodes:				a node buffer of size width by height;
 *		width, height:		dimensions of the node buffer;
 *		freePtX, freePtY:	an input node position;
 *	OUTPUT:
 *		insert a list of nodes along the minimum cost path from the seed node to the input node.  
 *		Notice that the seed node in the buffer has a NULL predecessor.
 *		And you want to insert pointer to Node in path, e.g., 
 *		insert nodes+j*width+i if you want to insert node at (i,j), instead of nodes[nodes+j*width+i]!!!
 *		after the procedure, the seed should be the head of path and the input code should be the tail
 */

void MinimumPath(CTypedPtrDblList <Node> *path, int freePtX, int freePtY, Node *nodes, int width, int height)
{
	Node* cn = &nodes[NINDEX(width, freePtY, freePtX)];
	
	// Follow the previous node link back until we hit the seed node (the one with a null prevNode)
	while (cn->prevNode != NULL)
	{
		path->AddHead(cn);
		cn = cn->prevNode;
	}
}

/************************An Extra Credit***************************
 *SeedSnap:
 *	INPUT:
 *		img:				a RGB image buffer of size width by height;
 *		width, height:		dimensions of the image buffer;
 *		x,y:				an input seed position;
 *	OUTPUT:
 *		update the value of x,y to the closest edge based on local image information.
 */

void SeedSnap(int &x, int &y, unsigned char *img, int width, int height)
{
	printf("SeedSnap in iScissor.cpp: to be implemented for extra credits!\n");
}

//generate a cost graph from original image and node buffer with all the link costs;
void MakeCostGraph(unsigned char *costGraph, const Node *nodes, const unsigned char *img, int imgWidth, int imgHeight)
{
	int graphWidth = imgWidth * 3;
	int graphHeight = imgHeight * 3;
	int dgX = 3;
	int dgY = 3*graphWidth;

	int i,j;
	for (j=0;j<imgHeight;j++)
	{
		for (i=0;i<imgWidth;i++)
		{
			int nodeIndex = j*imgWidth+i;
			int imgIndex = 3*nodeIndex;
			int costIndex = 3*((3*j+1)*graphWidth+(3*i+1));
			
			const Node *node = nodes+nodeIndex;
			const unsigned char *pxl = img + imgIndex;
			unsigned char *cst = costGraph + costIndex;

			cst[0] = pxl[0];
			cst[1] = pxl[1];
			cst[2] = pxl[2];
			
			//r,g,b channels are grad info in seperate channels;				
			DigitizeCost(cst	   + dgX, node->linkCost[0]);
			DigitizeCost(cst - dgY + dgX, node->linkCost[1]);
			DigitizeCost(cst - dgY      , node->linkCost[2]);
			DigitizeCost(cst - dgY - dgX, node->linkCost[3]);
			DigitizeCost(cst	   - dgX, node->linkCost[4]);
			DigitizeCost(cst + dgY - dgX, node->linkCost[5]);
			DigitizeCost(cst + dgY	   ,  node->linkCost[6]);
			DigitizeCost(cst + dgY + dgX, node->linkCost[7]);
		}
	}
}
